/****************************************************************
 * tape-library 						*
 *								*
 * Library of tape i/o routines to possibily do remote access	*
 *								*
 * This library is supposed to handle file names like		*
 *		system:/dev/xxxx				*
 *	if no system: appears all calls act normally		*
 *								*
 * Routines proviced						*
 *	r_open(device, mode)		open device		*
 *	r_close(fd)			close device		*
 *	r_read(fd, buffer, size)	read a buffer		*
 *	r_write(fd, buffer, size)	write a buffer		*
 *	r_ioctl(fd, func, parms)	do an ioctl		*
 *								*
 * Work to be done						*
 *	improve disconnect message				*
 ****************************************************************/

/****************************************************************
 * implemetation grunge.					*
 *								*
 *	The remote file stuff is done with /etc/rmt through	*
 *	/usr/ucb/rsh.  Since we need to know two file 		*
 *	descriptors for each one we are given by the user,	*
 *	we create a phony descripor by setting the R_BIT	*
 *	in the fd						*
 ****************************************************************/
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <stdio.h>
#include <sys/mtio.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>	

#include "t-lib.h"

#define MAX_REMOTES 2
/* fd's to communicate through the network */
static int	remote_pipe[MAX_REMOTES][4];
/* how each of the fd is used */
#define FROM_HOST 0 
#define TO_REMOTE 1
/* -- */
#define FROM_REMOTE 2
#define TO_HOST 3

static int r_reply(int rfd);

#define DUMMY 0	/* Dummy arguements */
/****************************************************************
 *								*
 * 	host (write)----->TO_REMOTE  == FROM_HOST <----- remote *
 * 	host (read)---->FROM_REMOTE  == TO_HOST <----- remote   *
 *								*
 ****************************************************************/
/* true if the fd is alread used */
static int	used[MAX_REMOTES] = {0, 0};

#define TRUE 1
#define FALSE 0
static int close_flag = FALSE;	/* ignore errors on close flag */

#define R_BIT  0x40

/****************************************************************
 * dis_connect -- print nasty message when we go bye bye	*
 ****************************************************************/
static void dis_connect(void)
{
    if (close_flag)
	return;		/* ignore errors on close */

    fprintf(stderr,"Disconnect error\n");
    perror("connection error:");
#ifdef DEBUG
    abort();
#endif DEBUG
    r_exit (8);
}
/****************************************************************
 * r_send_head -- send the heading to remote (no waiting for 	*
 *				the reply)			*
 *								*
 * Parameters							*
 *	rfd -- where to send it					*
 *	cmd_str -- printf string to send			*
 *	parm1, parm2 -- stuff for printf string			*
 ****************************************************************/
static void r_send_head(int rfd, char *cmd_str, int parm1, int parm2)
{
    char	full_cmd[100];	/* Complete command */
    int		full_len;	/* Length of complete command */

    sprintf(full_cmd, cmd_str, parm1, parm2);
    full_len = strlen(full_cmd);
    if (write(remote_pipe[rfd][TO_REMOTE], full_cmd, full_len) != full_len)
	dis_connect();
}
/****************************************************************
 * r_send -- send a command to the remote system		*
 *		and await a reply				*
 *								*
 * Parameters							*
 *	rfd -- where to send it					*
 *	cmd_str -- printf string to send			*
 *	parm1, parm2 -- stuff for printf string			*
 *								*
 * Returns							*
 *	response code from the other side			*
 ****************************************************************/
static int r_send(int rfd, char *cmd_str, int parm1, int parm2)
{
    r_send_head(rfd, cmd_str, parm1, parm2);
    return (r_reply(rfd));
}

/****************************************************************
 * r_close -- close the remote file				*
 *								*
 * Parameters							*
 *	fd -- the fd or phony fd to use	for the close		*
 ****************************************************************/
int r_close(int fd)
{
    int rfd;		/* remote file descriptior */
    int result;		/* result of the close */

    if ((fd & R_BIT) == 0)
	return (close(fd));

    rfd = fd & (~R_BIT);

    if (used[rfd] == 0)
	return (EBADF);

    close_flag = TRUE;
    result = r_send(rfd, "C\n", DUMMY, DUMMY);
    r_send_head(rfd,"X\nX\nX\nX\nX\n", DUMMY, DUMMY);/* force an exit? */
    close_flag = FALSE;

    if (killpg(used[rfd], 7) != 0)		/* get rid of child */
	fprintf(stderr,"Kill failed \n");

    used[rfd] = 0;
    close(remote_pipe[rfd][0]);
    close(remote_pipe[rfd][1]);
    close(remote_pipe[rfd][2]);
    close(remote_pipe[rfd][3]);
    return (result);
}

/****************************************************************
 * r_read -- read data from the remote fd			*
 *								*
 * Parameters							*
 *	exactly like the read system call			*
 ****************************************************************/
int r_read(int fd, char *buffer, int size)
{
    int rfd;		/* remote file descriptor */
    int count;		/* number of bytes current done */
    int act_size;		/* size of expected read */
    int read_size;		/* size of last read */

    if ((fd & R_BIT) == 0)
	return (read(fd, buffer, size));
    
    rfd = fd & (~R_BIT);
    act_size = r_send(rfd, "R%d\n", size, DUMMY);

    if (act_size <= 0)
	return (act_size);

    for (count = 0; count < act_size; count += read_size) {
	read_size = read(remote_pipe[rfd][FROM_REMOTE], 
			buffer+count, act_size - count);
	if (read_size < 0)
	    dis_connect();
    }
    return (act_size);
}

/****************************************************************
 * r_gets -- get a string from the other side			*
 *								*
 * Parameter							*
 *	rfd -- remote fd to use					*
 *	str -- place to stuff the result string			*
 ****************************************************************/
static void r_gets(int rfd, char *str)
{
    while (1) {
	if (read(remote_pipe[rfd][FROM_REMOTE], str, 1) != 1)
	    dis_connect();

	if ((*str) == '\n') {
	    return;
	}
	str++;
    }
}

/****************************************************************
 * r_reply -- read the reply from the remote			*
 *								*
 * Paremters							*
 *	rfd -- the fd to use for commuications			*
 *								*
 * Returns							*
 *	Result code from the other side or -1 for error		*
 ****************************************************************/
static int r_reply(int rfd)
{
    char	reply[100];	/* repsonce to message */
    int		r_errno;	/* local copy of error */

    memset(reply, '\0', sizeof(reply));
    r_gets(rfd, reply);

    switch (reply[0]) {
	case 'E':
	    r_errno = atoi(reply+1);
	    memset(reply, '\0', sizeof(reply));
	    r_gets(rfd, reply);
	    errno = r_errno;
	    return (-1);

	case 'A':
	    return (atoi(reply+1));

	default:
	    fprintf(stderr,"Rmt:%d protocall botch\n", rfd);
	    dis_connect();
	    /*NOTREACHED*/
    }
    return (0);
}

/****************************************************************
 * r_write -- write to fd or remote				*
 *								*
 * Parameters							*
 *	just like system call					*
 ****************************************************************/
int r_write(int fd, char *buffer, int size)
{
    int rfd;		/* remote file descriptor */

    if ((fd & R_BIT) == 0)
	return (write(fd, buffer, size));
    
    rfd = fd & (~R_BIT);

    r_send_head(rfd, "W%d\n", size, DUMMY);

    if (write(remote_pipe[rfd][TO_REMOTE], buffer, size) != size)
	dis_connect();

    return (r_reply(rfd));
}
/****************************************************************
 * r_ioctl -- do an ioctl on the device				*
 *								*
 * Parameters							*
 *	same as system call					*
 ****************************************************************/
int r_ioctl(int fd, int funct, struct mtop *mtop)
{
    int rfd;		/* remote file descriptor */

    if ((fd & R_BIT) == 0)
	return (ioctl(fd, funct, (char *)mtop));
    
    rfd = fd & (~R_BIT);
    return (r_send(rfd, "I%d\n%d\n", mtop->mt_op, mtop->mt_count));
}

/****************************************************************
 * r_exit -- just like exit, but kill all remotes		*
 ****************************************************************/
void r_exit(int code)
{
    int rfd;

    for (rfd = 0; rfd < MAX_REMOTES; rfd++)  {
	if (used[rfd] != 0)
	    killpg(used[rfd], 7);
    }
    
    exit(code);
}

/****************************************************************
 * r_open -- open a file					*
 *								*
 * Parameters							*
 *	name -- the device name with possible system: added	*
 *	mode -- mode to use for the open			*
 *								*
 * Returns							*
 *	fd of open file						*
 *	funny fd of the open file on remote machine		*
 *	-1 for error (errno set)				*
 ****************************************************************/
int r_open(char *name, int mode)
{
    char	host_name[MAXPATHLEN];	/* the host name */
    char	file_name[MAXPATHLEN];	/* the file name only */
    char	*colon_ptr;		/* where the : is */
    int		rfd;			/* remote file descriptior */

    close_flag = FALSE;

    colon_ptr = strchr(name, ':');
    if (colon_ptr == NULL) 
	return (open(name, mode));

    /* break apart the name */
    strcpy(host_name, name);
    (*strchr(host_name, ':')) = '\0';
    strcpy(file_name, colon_ptr+1);

    if (gethostbyname(host_name) == NULL) {
	errno = EHOSTUNREACH;
	return (-1);
    }

    for (rfd = 0; rfd < MAX_REMOTES; rfd++) {
	if (used[rfd] == 0) {
	    pipe(&remote_pipe[rfd][0]);
	    pipe(&remote_pipe[rfd][2]);
	    used[rfd] = fork();
	    if (used[rfd] < 0) {
		fprintf(stderr,"Fork failed\n");
		perror("r-tape:");
		exit (8);
	    }
	    /* check for child */
	    if (used[rfd] == 0) {
		int i;

		setpgrp();
		dup2(remote_pipe[rfd][FROM_HOST], 0);
		dup2(remote_pipe[rfd][TO_HOST], 1);
		dup2(remote_pipe[rfd][TO_HOST], 2);

		for (i = 3; i < NOFILE; i++)
		    close(i);

		execl("/usr/ucb/rsh", "rsh",  host_name, "/etc/rmt", 0);
		exit (8);
	    }

	    if (r_send(rfd, "O%s\n%d\n", (int)file_name, mode) < 0) {

		/* get rid of child */
		if (killpg(used[rfd], 7) != 0)		
		    fprintf(stderr,"Kill failed \n");

		used[rfd] = 0;
		return (-1);
	    }
	    return (rfd | R_BIT);
	}
    }
    errno = ENFILE;
    return (-1);
}
